<?php
// $Id: file.inc,v 1.1.2.4 2009/07/22 23:31:39 doq Exp $

/**
 * File cache engine.
 */
class fileCacheEngine extends CacheEngine {
  var $fspath = '/tmp/filecache';

  /**
   * @return
   *   TRUE if cache engine is available; FALSE otherwise.
   */
  function status() {
    return TRUE;
  }

  function page_fast_cache() {
    return $this->fast_cache;
  }

  function __construct($settings) {
    parent::__construct($settings);

    if (isset($this->settings['path'])) {
      $this->fspath = $this->settings['path'];
    }
  }

  
  function get($key) {
//    global $user;
    
//    $cache_lifetime = variable_get('cache_lifetime', 0);

//    if (variable_get('page_cache_fastpath', 0)) {
    $cache = NULL;
    $cache_file = $this->key($key);

    if (file_exists($cache_file)) {
      if ($fp = fopen($cache_file, 'r')) {
        if (flock($fp, LOCK_SH)) {
          $data = fread($fp, filesize($cache_file));
          assert(strlen($data) == filesize($cache_file));
          flock($fp, LOCK_UN);
          $cache = unserialize($data);
        }
//        chmod($cache_file, 0664); // Necessary for non-webserver users.
        fclose($fp);
      }
    }
    return $cache;
/*
    }
    else {
      // garbage collection necessary when enforcing a minimum cache lifetime
      $cache_flush = variable_get('cache_flush', 0);
      if ($cache_flush && ($cache_flush + $cache_lifetime <= time())) {
        variable_set('cache_flush', 0);
      }
    }

    if (isset($cache->data)) {
      // If enforcing a minimum cache lifetime, validate that the data is
      // currently valid for this user before we return it by making sure the
      // cache entry was created before the timestamp in the current session's
      // cache timer. The cache variable is loaded into the $user object by
      // sess_read() in session.inc.
      if ($user->cache > $cache->created) {
        // This cache data is too old and thus not valid for us, ignore it.
        return FALSE;
      }
      return $cache;
    }
    return FALSE;
*/
  }
  
  function set($key, $value, $expire = CACHE_PERMANENT, $headers = NULL) {
    static $subdirectories;

//    if (variable_get('page_cache_fastpath', 0)) {
    // prepare the cache before grabbing the file lock
    $cache = new stdClass;
    $cache->cid = $key;
//    $cache->table = $this->name;
    $cache->created = time();
    $cache->expire = $expire;
    $cache->headers = $headers;
    $cache->data = $value;
    $cache->serialized = FALSE;

    $data = serialize($cache);

    assert(strlen($key) > 0);

    $file = $this->key($key);
    if ($fp = @fopen($file, 'a')) {
      // Only write to the cache file if we can obtain an exclusive lock.
      if (flock($fp, LOCK_EX)) {
        ftruncate($fp, 0);
        fwrite($fp, $data);
        flock($fp, LOCK_UN);
      }
      fclose($fp);

//      touch($file, time()); ///////////////

      chmod($file, 0664); // Necessary for non-webserver users.

      if ($expire == CACHE_PERMANENT) {
        touch($file, 1000);
      }
    }
    else {
      watchdog('cache', 'Cache write error, failed to open file %file.', array('%file' => $file), WATCHDOG_ERROR);
    }
//    }
  }
  
  function delete($key, $wildcard) {
    // Remove from static array cache.
    parent::delete($key, $wildcard);

    if ($wildcard) {
      if ($key == '*') {
        $fspath = $this->fspath;
        $files = file_scan_directory($fspath .'/'. $this->name, '.*', array('.', '..', 'CVS')/*, 0, FALSE*/);
        foreach ($files as $file) {
          if ($fp = fopen($file->filename, 'a')) {
            // only delete the cache file once we obtain an exclusive lock to prevent
            // deleting a cache file that is currently being read.
            if (flock($fp, LOCK_EX)) {
              unlink($file->filename);
            }
          }
        }
      }
      else {
//        $look_for = explode('*', $key);
        $fspath = $this->fspath;
        // Filename: abcdef12345verylongmd5code--content:123456:987654
        /// @todo Do we need to add $look_for[0] = $this->key($key)  ?
        $files = file_scan_directory($fspath .'/'. $this->name, ".--$key.", array('.', '..', 'CVS')/*, 0, FALSE*/);
        foreach ($files as $file) {
          if ($fp = fopen($file->filename, 'a')) {
            // Only delete the cache file once we obtain an exclusive lock to prevent
            // deleting a cache file that is currently being read.
            if (flock($fp, LOCK_EX)) {
              unlink($file->filename);
            }
          }
        }
      }
    }
    else {
      $file = $this->key($key);
      if (file_exists($file) && ($fp = fopen($file, 'a'))) {
        // only delete the cache file once we obtain an exclusive lock to prevent
        // deleting a cache file that is currently being read.
        if (flock($fp, LOCK_EX)) {
          unlink($file);
        }
      }
    }

  }
  
  function flush($flush = 0) {
    $flush = !$flush ? time() : $flush;
//    global $user;
    
    $table = $this->name;
//    $cache_lifetime = variable_get('cache_lifetime', 0);
    $fspath = $this->fspath;
    
    $ivalfrom = ord("a");
    $ivalto = ord("f");
    
    for ($i = $ivalfrom; $i <= $ivalto; ++$i) {
      $this->purge("$fspath/$table/". chr($i), $flush);
    }

    for ($i = 0; $i < 10; ++$i) {
      $this->purge("$fspath/$table/". $i, $flush);
    }
  }
  
  function key($key) {
    $table = $this->name;
    $fspath = $this->fspath;
    $hash = md5($key);
    /// @todo make sure we always get valid filenames in the appendix.
    $appendix = str_replace(array('/'), array('-'), $key);
    
    $this->create_directory($fspath, $hash{0});
    
    return "$fspath/$table/". $hash{0}. '/'. $hash. '--'. $appendix;
  }
  
  /**
   * Create the necessary $table directory and/or letter/number subdirectory if
   * it doesn't exist.  We store the directories we've created in a static 
   * so we don't bother doing an fstat on that more than one time per page load.
   */
  function create_directory($fspath, $hash) {
    static $dirs = array();
    $table = $this->name;
    
    $create = array($fspath, "$fspath/$table", "$fspath/$table/$hash");

    foreach ($create as $dir) {
      $dir = rtrim($dir, '/\\');
      // Check the static $dirs to avoid excessive fstats
      if (!isset($dirs[$dir])) {
        $dirs[$dir] = 1;
        if (!is_dir($dir)) {
          if (!mkdir($dir)) {
            watchdog('cache', 'Failed to create directory %dir.', array('%dir' => $dir), WATCHDOG_ERROR);
          }
          else {
            @chmod($dir, 0775); // Necessary for non-webserver users.
          }
        }
      }
    }
  }

  function purge($dir, $flush) {
    //global $cache_lifetime;
    //$cache_lifetime = variable_get('cache_lifetime', 0);
    $files = file_scan_directory($dir, '.*', array('.', '..', 'CVS'), 0, FALSE);
    foreach ($files as $file) {
      $mtime = filemtime($file->filename);
      if ($mtime != 1000 && $mtime < ($flush/* - $cache_lifetime*/)) {
        if ($fp = fopen($file->filename, 'r')) {
          // We need an exclusive lock, but don't block if we can't get it as
          // we can simply try again next time cron is run.
          if (flock($fp, LOCK_EX|LOCK_NB)) {
            unlink($file->filename);
          }
        }
      }
    }
  }

  /**
   * Statistics information.
   */
  function stats() {
    $stats = array(
      'uptime' => time(),
      'bytes_used' => disk_total_space($this->fspath) - disk_free_space($this->fspath),
      'bytes_total' => disk_total_space($this->fspath),
      'gets' => 0,
      'sets' => 0,
      'hits' => 0,
      'misses' => 0,
      'req_rate' => 0,
      'hit_rate' => 0,
      'miss_rate' => 0,
      'set_rate' => 0,
    );
    return $stats;
  }
}
